# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("opengl32.dll")
prototypes = \
    {
        # 
        'wglSwapMultipleBuffers': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"hdc": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "uiFlags": SimTypeInt(signed=False, label="UInt32")}, name="WGLSWAP", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["param0", "param1"]),
        # 
        'wglCopyContext': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2"]),
        # 
        'wglCreateContext': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["param0"]),
        # 
        'wglCreateLayerContext': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["param0", "param1"]),
        # 
        'wglDeleteContext': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        # 
        'wglGetCurrentContext': SimTypeFunction([], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)),
        # 
        'wglGetCurrentDC': SimTypeFunction([], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)),
        # 
        'wglGetProcAddress': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypePointer(SimTypeFunction([], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)), offset=0), arg_names=["param0"]),
        # 
        'wglMakeCurrent': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1"]),
        # 
        'wglShareLists': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1"]),
        # 
        'wglUseFontBitmapsA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]),
        # 
        'wglUseFontBitmapsW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]),
        # 
        'wglUseFontOutlinesA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeFloat(size=32), SimTypeFloat(size=32), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"gmfBlackBoxX": SimTypeFloat(size=32), "gmfBlackBoxY": SimTypeFloat(size=32), "gmfptGlyphOrigin": SimStruct({"x": SimTypeFloat(size=32), "y": SimTypeFloat(size=32)}, name="POINTFLOAT", pack=False, align=None), "gmfCellIncX": SimTypeFloat(size=32), "gmfCellIncY": SimTypeFloat(size=32)}, name="GLYPHMETRICSFLOAT", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3", "param4", "param5", "param6", "param7"]),
        # 
        'wglUseFontOutlinesW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeFloat(size=32), SimTypeFloat(size=32), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"gmfBlackBoxX": SimTypeFloat(size=32), "gmfBlackBoxY": SimTypeFloat(size=32), "gmfptGlyphOrigin": SimStruct({"x": SimTypeFloat(size=32), "y": SimTypeFloat(size=32)}, name="POINTFLOAT", pack=False, align=None), "gmfCellIncX": SimTypeFloat(size=32), "gmfCellIncY": SimTypeFloat(size=32)}, name="GLYPHMETRICSFLOAT", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3", "param4", "param5", "param6", "param7"]),
        # 
        'wglDescribeLayerPlane': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"nSize": SimTypeShort(signed=False, label="UInt16"), "nVersion": SimTypeShort(signed=False, label="UInt16"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "iPixelType": SimTypeChar(label="Byte"), "cColorBits": SimTypeChar(label="Byte"), "cRedBits": SimTypeChar(label="Byte"), "cRedShift": SimTypeChar(label="Byte"), "cGreenBits": SimTypeChar(label="Byte"), "cGreenShift": SimTypeChar(label="Byte"), "cBlueBits": SimTypeChar(label="Byte"), "cBlueShift": SimTypeChar(label="Byte"), "cAlphaBits": SimTypeChar(label="Byte"), "cAlphaShift": SimTypeChar(label="Byte"), "cAccumBits": SimTypeChar(label="Byte"), "cAccumRedBits": SimTypeChar(label="Byte"), "cAccumGreenBits": SimTypeChar(label="Byte"), "cAccumBlueBits": SimTypeChar(label="Byte"), "cAccumAlphaBits": SimTypeChar(label="Byte"), "cDepthBits": SimTypeChar(label="Byte"), "cStencilBits": SimTypeChar(label="Byte"), "cAuxBuffers": SimTypeChar(label="Byte"), "iLayerPlane": SimTypeChar(label="Byte"), "bReserved": SimTypeChar(label="Byte"), "crTransparent": SimTypeInt(signed=False, label="UInt32")}, name="LAYERPLANEDESCRIPTOR", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3", "param4"]),
        # 
        'wglSetLayerPaletteEntries': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3", "param4"]),
        # 
        'wglGetLayerPaletteEntries': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3", "param4"]),
        # 
        'wglRealizeLayerPalette': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2"]),
        # 
        'wglSwapLayerBuffers': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1"]),
    }

lib.set_prototypes(prototypes)



from ailment.expression import BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class EagerEvaluation(PeepholeOptimizationExprBase):
    """
    Eagerly evaluates certain types of expressions.
    """
    __slots__ = ()

    name = "Eager expression evaluation"
    expr_classes = (BinaryOp, )

    def optimize(self, expr: BinaryOp):

        if expr.op == "Add" \
                and isinstance(expr.operands[0], Const) \
                and isinstance(expr.operands[1], Const):
            mask = (2 << expr.bits) - 1
            new_expr = Const(expr.idx, None,
                             (expr.operands[0].value + expr.operands[1].value) & mask,
                             expr.bits,
                             **expr.tags)
            return new_expr

        return None



# pylint:disable=import-outside-toplevel
from functools import wraps
from typing import Optional, List

binsync_available = None
binsync = None
Client = None
StackVariable = None
StackOffsetType = None



from ... import knowledge_plugins
from ...knowledge_plugins.plugin import KnowledgeBasePlugin
from ...sim_variable import SimStackVariable
from ..variables.variable_manager import VariableManagerInternal


def import_binsync():
    global binsync_available
    try:
        import binsync
        from binsync.client import Client
        from binsync.data.stack_variable import StackVariable, StackOffsetType
        binsync_available = True
    except ImportError:
        binsync_available = False


def make_state(f):
    """
    Build a writeable State instance and pass to `f` as the `state` kwarg if the `state` kwarg is None.
    Function `f` should have have at least two kwargs, `user` and `state`.
    """

    @wraps(f)
    def state_check(self, *args, **kwargs):
        state = kwargs.pop('state', None)
        user = kwargs.pop('user', None)
        if state is None:
            state = self.client.get_state(user=user)
            kwargs['state'] = state
            r = f(self, *args, **kwargs)
            state.save()
            return r
        else:
            kwargs['state'] = state
            r = f(self, *args, **kwargs)
            return r

    return state_check


def make_ro_state(f):
    """
    Build a read-only State instance and pass to `f` as the `state` kwarg if the `state` kwarg is None.
    Function `f` should have have at least two kwargs, `user` and `state`.
    """

    @wraps(f)
    def state_check(self, *args, **kwargs):
        state = kwargs.pop('state', None)
        user = kwargs.pop('user', None)
        if state is None:
            state = self.client.get_state(user=user)
        kwargs['state'] = state
        kwargs['user'] = user
        return f(self, *args, **kwargs)

    return state_check


def init_checker(f):
    @wraps(f)
    def initcheck(self, *args, **kwargs):
        if self.client is None:
            raise ValueError("Please initialize SyncController by calling initialize(client).")
        return f(self, *args, **kwargs)
    return initcheck


class SyncController(KnowledgeBasePlugin):
    """
    SyncController interfaces with a binsync client to push changes upwards and pull changes downwards.

    :ivar binsync.Client client:   The binsync client.
    """
    def __init__(self, kb):

        # import binsync upon the first use of this class
        import_binsync()

        super().__init__()

        self._kb: KnowledgeBasePlugin = kb
        self.client: Optional['binsync.client.Client'] = None

    #
    # Public methods
    #

    def connect(self, user, path,
                bin_hash="", init_repo=False, ssh_agent_pid=None, ssh_auth_sock=None, remote_url=None):
        self.client = Client(user, path, bin_hash,
                             init_repo=init_repo,
                             ssh_agent_pid=ssh_agent_pid,
                             ssh_auth_sock=ssh_auth_sock,
                             remote_url=remote_url)

    @property
    def connected(self):
        return self.client is not None

    def commit(self):
        self.client.save_state()

    def update(self):
        self.client.update()

    def copy(self):
        raise NotImplementedError

    def pull(self):
        return self.client.pull()

    @property
    def has_remote(self):
        return self.client.has_remote

    @init_checker
    def users(self):
        return self.client.users()

    @init_checker
    def status(self):
        return self.client.status()

    def tally(self, users=None):
        return self.client.tally(users=users)

    #
    # Pushers
    #

    @init_checker
    @make_state
    # pylint:disable=unused-argument,no-self-use
    def push_function(self, func: knowledge_plugins.Function, user=None, state=None):
        """
        Push a function upwards.

        :param Function func:   The angr Function object to push upwards.
        :return:                True if updates are made. False otherwise.
        :rtype:                 bool
        """

        _func = binsync.data.Function(func.addr, name=func.name, notes=None)
        return state.set_function(_func)

    @init_checker
    @make_state
    # pylint:disable=unused-argument,no-self-use
    def push_comment(self, addr, comment, decompiled=False, user=None, state=None):
        func_addr = self.get_func_addr_from_addr(addr)
        sync_cmt = binsync.data.Comment(func_addr, addr, comment, decompiled)

        return state.set_comment(sync_cmt)

    @init_checker
    @make_state
    # pylint:disable=unused-argument,no-self-use
    def push_comments(self, comments: List['binsync.data.Comment'], user=None, state=None):
        """
        Push a bunch of comments upwards.

        :param list comments:   A list of BinSync Comments
        :return:                bool
        """

        r = True
        for cmt in comments:
            r &= state.set_comment(cmt)
        return r

    @init_checker
    @make_state
    # pylint:disable=unused-argument,no-self-use
    def push_stack_variables(self, stack_variables: List[SimStackVariable], var_manager: VariableManagerInternal,
                             user=None, state=None):
        """

        :param stack_variables:
        :param var_manager:
        :return:
        """
        r = True
        for var in stack_variables:
            guessed_var_type = var_manager.get_variable_type(var)
            var_type = guessed_var_type if guessed_var_type else "BOT"

            # construct a StackVariable for each SimStackVariable
            sync_stack_var = StackVariable(var.offset, StackOffsetType.ANGR, var.name,
                                           var_type, var.size, var_manager.func_addr)

            r &= state.set_stack_variable(sync_stack_var, var.offset, var_manager.func_addr)

        # return true only if all pushed worked
        return r

    @init_checker
    @make_state
    # pylint:disable=unused-argument,no-self-use
    def push_stack_variable(self, func_addr, offset, name, type_, size_, user=None, state=None):
        sync_var = StackVariable(offset, StackOffsetType.ANGR, name, type_, size_, func_addr)
        return state.set_stack_variable(sync_var, offset, func_addr)

    #
    # Pullers
    #

    @init_checker
    @make_ro_state
    # pylint:disable=unused-argument
    def pull_function(self, addr, user=None, state=None) -> Optional['binsync.data.Function']:
        """
        Pull a function downwards.

        :param int addr:    Address of the function.
        :param str user:    Name of the user.
        :return:            The binsync.data.Function object if pulling succeeds, or None if pulling fails.
        """

        func_addr = self.get_func_addr_from_addr(addr)
        try:
            func = state.get_function(func_addr)
            return func
        except KeyError:
            return None

    @init_checker
    @make_ro_state
    # pylint:disable=unused-argument
    def pull_comment(self, addr, user=None, state=None) -> Optional['binsync.data.Comment']:
        """
        Pull a comment downwards.

        :param int addr:    Address of the comment.
        :param str user:    Name of the user.
        :return:            a Comment object from BinSync, or None
        """

        func_addr = self.get_func_addr_from_addr(addr)
        try:
            comment = state.get_comment(func_addr, addr)
            return comment
        except KeyError:
            return None

    @init_checker
    @make_ro_state
    # pylint:disable=unused-argument,no-self-use
    def pull_comments(self, func_addr, user=None, state=None):
        """
        Pull comments downwards.

        :param int start_addr:  Where we want to pull comments.
        :param int end_addr:    Where we want to stop pulling comments (exclusive).
        :return:                An iterator.
        :rtype:                 Iterable
        """
        try:
            comments = state.get_comments(func_addr)
            return comments
        except KeyError:
            return {}

    @init_checker
    @make_ro_state
    # pylint:disable=unused-argument,no-self-use
    def pull_patches(self, user=None, state=None):
        """
        Pull patches.

        :param str user:    Name of the user to patches from.
        :return:            An iterator
        :rtype:             Iterable
        """

        return state.get_state(user=user).get_patches()

    @init_checker
    @make_ro_state
    # pylint:disable=unused-argument,no-self-use
    def pull_stack_variables(self, func_addr, user=None, state=None):
        """
        Pull stack variables from a function.

        @param func_addr:   Function address to pull from
        @param user:
        @param state:
        @return:
        """
        try:
            stack_vars = state.get_stack_variables(func_addr)
        except KeyError:
            stack_vars = {}

        return stack_vars

    #
    #   Utils
    #

    def get_func_addr_from_addr(self, addr):
        try:
            func_addr = self._kb.cfgs.get_most_accurate().get_any_node(addr, anyaddr=True).function_address
        except AttributeError:
            func_addr = -1

        return func_addr


if binsync_available:
    KnowledgeBasePlugin.register_default("sync", SyncController)



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_x86_LE_32_System_Management_Mode(ArchPcode):
    name = 'x86:LE:32:System Management Mode'
    pcode_arch = 'x86:LE:32:System Management Mode'
    description = 'Intel/AMD 32-bit x86 System Management Mode'
    bits = 32
    ip_offset = 0x284
    sp_offset = 0x10
    bp_offset = sp_offset
    instruction_endness = Endness.LE
    register_list = [
        Register('eax', 4, 0x0),
        Register('ax', 2, 0x0),
        Register('al', 1, 0x0),
        Register('ah', 1, 0x1),
        Register('ecx', 4, 0x4),
        Register('cx', 2, 0x4),
        Register('cl', 1, 0x4),
        Register('ch', 1, 0x5),
        Register('edx', 4, 0x8),
        Register('dx', 2, 0x8),
        Register('dl', 1, 0x8),
        Register('dh', 1, 0x9),
        Register('ebx', 4, 0xc),
        Register('bx', 2, 0xc),
        Register('bl', 1, 0xc),
        Register('bh', 1, 0xd),
        Register('esp', 4, 0x10),
        Register('sp', 2, 0x10),
        Register('ebp', 4, 0x14),
        Register('bp', 2, 0x14),
        Register('esi', 4, 0x18),
        Register('si', 2, 0x18),
        Register('edi', 4, 0x1c),
        Register('di', 2, 0x1c),
        Register('es', 2, 0x100),
        Register('cs', 2, 0x102),
        Register('ss', 2, 0x104),
        Register('ds', 2, 0x106),
        Register('fs', 2, 0x108),
        Register('gs', 2, 0x10a),
        Register('fs_offset', 4, 0x110),
        Register('gs_offset', 4, 0x114),
        Register('cf', 1, 0x200),
        Register('f1', 1, 0x201),
        Register('pf', 1, 0x202),
        Register('f3', 1, 0x203),
        Register('af', 1, 0x204),
        Register('f5', 1, 0x205),
        Register('zf', 1, 0x206),
        Register('sf', 1, 0x207),
        Register('tf', 1, 0x208),
        Register('if', 1, 0x209),
        Register('df', 1, 0x20a),
        Register('of', 1, 0x20b),
        Register('iopl', 1, 0x20c),
        Register('nt', 1, 0x20d),
        Register('f15', 1, 0x20e),
        Register('rf', 1, 0x20f),
        Register('vm', 1, 0x210),
        Register('ac', 1, 0x211),
        Register('vif', 1, 0x212),
        Register('vip', 1, 0x213),
        Register('id', 1, 0x214),
        Register('eflags', 4, 0x280),
        Register('flags', 2, 0x280),
        Register('eip', 4, 0x284, alias_names=('pc', 'ip')),
        Register('ip', 2, 0x284),
        Register('dr0', 4, 0x300),
        Register('dr1', 4, 0x304),
        Register('dr2', 4, 0x308),
        Register('dr3', 4, 0x30c),
        Register('dr4', 4, 0x310),
        Register('dr5', 4, 0x314),
        Register('dr6', 4, 0x318),
        Register('dr7', 4, 0x31c),
        Register('cr0', 4, 0x320),
        Register('cr2', 4, 0x328),
        Register('cr3', 4, 0x32c),
        Register('cr4', 4, 0x330),
        Register('tr0', 4, 0x400),
        Register('tr1', 4, 0x404),
        Register('tr2', 4, 0x408),
        Register('tr3', 4, 0x40c),
        Register('tr4', 4, 0x410),
        Register('tr5', 4, 0x414),
        Register('tr6', 4, 0x418),
        Register('tr7', 4, 0x41c),
        Register('xcr0', 8, 0x600),
        Register('bndcfgs', 8, 0x700),
        Register('bndcfgu', 8, 0x708),
        Register('bndstatus', 8, 0x710),
        Register('bnd0', 16, 0x740),
        Register('bnd0_lb', 8, 0x740),
        Register('bnd0_ub', 8, 0x748),
        Register('bnd1', 16, 0x750),
        Register('bnd1_lb', 8, 0x750),
        Register('bnd1_ub', 8, 0x758),
        Register('bnd2', 16, 0x760),
        Register('bnd2_lb', 8, 0x760),
        Register('bnd2_ub', 8, 0x768),
        Register('bnd3', 16, 0x770),
        Register('bnd3_lb', 8, 0x770),
        Register('bnd3_ub', 8, 0x778),
        Register('ssp', 8, 0x7c0),
        Register('ia32_pl2_ssp', 8, 0x7c8),
        Register('ia32_pl1_ssp', 8, 0x7d0),
        Register('ia32_pl0_ssp', 8, 0x7d8),
        Register('st0', 10, 0x1000),
        Register('st1', 10, 0x100a),
        Register('st2', 10, 0x1014),
        Register('st3', 10, 0x101e),
        Register('st4', 10, 0x1028),
        Register('st5', 10, 0x1032),
        Register('st6', 10, 0x103c),
        Register('st7', 10, 0x1046),
        Register('c0', 1, 0x1080),
        Register('c1', 1, 0x1081),
        Register('c2', 1, 0x1082),
        Register('c3', 1, 0x1083),
        Register('mxcsr', 4, 0x1084),
        Register('fpucontrolword', 2, 0x1090),
        Register('fpustatusword', 2, 0x1092),
        Register('fputagword', 2, 0x1094),
        Register('fpulastinstructionopcode', 2, 0x1096),
        Register('fpudatapointer', 4, 0x1098),
        Register('fpuinstructionpointer', 4, 0x109c),
        Register('mm0', 8, 0x1100),
        Register('mm0_da', 4, 0x1100),
        Register('mm0_wa', 2, 0x1100),
        Register('mm0_ba', 1, 0x1100),
        Register('mm0_bb', 1, 0x1101),
        Register('mm0_wb', 2, 0x1102),
        Register('mm0_bc', 1, 0x1102),
        Register('mm0_bd', 1, 0x1103),
        Register('mm0_db', 4, 0x1104),
        Register('mm0_wc', 2, 0x1104),
        Register('mm0_be', 1, 0x1104),
        Register('mm0_bf', 1, 0x1105),
        Register('mm0_wd', 2, 0x1106),
        Register('mm0_bg', 1, 0x1106),
        Register('mm0_bh', 1, 0x1107),
        Register('mm1', 8, 0x1108),
        Register('mm1_da', 4, 0x1108),
        Register('mm1_wa', 2, 0x1108),
        Register('mm1_ba', 1, 0x1108),
        Register('mm1_bb', 1, 0x1109),
        Register('mm1_wb', 2, 0x110a),
        Register('mm1_bc', 1, 0x110a),
        Register('mm1_bd', 1, 0x110b),
        Register('mm1_db', 4, 0x110c),
        Register('mm1_wc', 2, 0x110c),
        Register('mm1_be', 1, 0x110c),
        Register('mm1_bf', 1, 0x110d),
        Register('mm1_wd', 2, 0x110e),
        Register('mm1_bg', 1, 0x110e),
        Register('mm1_bh', 1, 0x110f),
        Register('mm2', 8, 0x1110),
        Register('mm2_da', 4, 0x1110),
        Register('mm2_wa', 2, 0x1110),
        Register('mm2_ba', 1, 0x1110),
        Register('mm2_bb', 1, 0x1111),
        Register('mm2_wb', 2, 0x1112),
        Register('mm2_bc', 1, 0x1112),
        Register('mm2_bd', 1, 0x1113),
        Register('mm2_db', 4, 0x1114),
        Register('mm2_wc', 2, 0x1114),
        Register('mm2_be', 1, 0x1114),
        Register('mm2_bf', 1, 0x1115),
        Register('mm2_wd', 2, 0x1116),
        Register('mm2_bg', 1, 0x1116),
        Register('mm2_bh', 1, 0x1117),
        Register('mm3', 8, 0x1118),
        Register('mm3_da', 4, 0x1118),
        Register('mm3_wa', 2, 0x1118),
        Register('mm3_ba', 1, 0x1118),
        Register('mm3_bb', 1, 0x1119),
        Register('mm3_wb', 2, 0x111a),
        Register('mm3_bc', 1, 0x111a),
        Register('mm3_bd', 1, 0x111b),
        Register('mm3_db', 4, 0x111c),
        Register('mm3_wc', 2, 0x111c),
        Register('mm3_be', 1, 0x111c),
        Register('mm3_bf', 1, 0x111d),
        Register('mm3_wd', 2, 0x111e),
        Register('mm3_bg', 1, 0x111e),
        Register('mm3_bh', 1, 0x111f),
        Register('mm4', 8, 0x1120),
        Register('mm4_da', 4, 0x1120),
        Register('mm4_wa', 2, 0x1120),
        Register('mm4_ba', 1, 0x1120),
        Register('mm4_bb', 1, 0x1121),
        Register('mm4_wb', 2, 0x1122),
        Register('mm4_bc', 1, 0x1122),
        Register('mm4_bd', 1, 0x1123),
        Register('mm4_db', 4, 0x1124),
        Register('mm4_wc', 2, 0x1124),
        Register('mm4_be', 1, 0x1124),
        Register('mm4_bf', 1, 0x1125),
        Register('mm4_wd', 2, 0x1126),
        Register('mm4_bg', 1, 0x1126),
        Register('mm4_bh', 1, 0x1127),
        Register('mm5', 8, 0x1128),
        Register('mm5_da', 4, 0x1128),
        Register('mm5_wa', 2, 0x1128),
        Register('mm5_ba', 1, 0x1128),
        Register('mm5_bb', 1, 0x1129),
        Register('mm5_wb', 2, 0x112a),
        Register('mm5_bc', 1, 0x112a),
        Register('mm5_bd', 1, 0x112b),
        Register('mm5_db', 4, 0x112c),
        Register('mm5_wc', 2, 0x112c),
        Register('mm5_be', 1, 0x112c),
        Register('mm5_bf', 1, 0x112d),
        Register('mm5_wd', 2, 0x112e),
        Register('mm5_bg', 1, 0x112e),
        Register('mm5_bh', 1, 0x112f),
        Register('mm6', 8, 0x1130),
        Register('mm6_da', 4, 0x1130),
        Register('mm6_wa', 2, 0x1130),
        Register('mm6_ba', 1, 0x1130),
        Register('mm6_bb', 1, 0x1131),
        Register('mm6_wb', 2, 0x1132),
        Register('mm6_bc', 1, 0x1132),
        Register('mm6_bd', 1, 0x1133),
        Register('mm6_db', 4, 0x1134),
        Register('mm6_wc', 2, 0x1134),
        Register('mm6_be', 1, 0x1134),
        Register('mm6_bf', 1, 0x1135),
        Register('mm6_wd', 2, 0x1136),
        Register('mm6_bg', 1, 0x1136),
        Register('mm6_bh', 1, 0x1137),
        Register('mm7', 8, 0x1138),
        Register('mm7_da', 4, 0x1138),
        Register('mm7_wa', 2, 0x1138),
        Register('mm7_ba', 1, 0x1138),
        Register('mm7_bb', 1, 0x1139),
        Register('mm7_wb', 2, 0x113a),
        Register('mm7_bc', 1, 0x113a),
        Register('mm7_bd', 1, 0x113b),
        Register('mm7_db', 4, 0x113c),
        Register('mm7_wc', 2, 0x113c),
        Register('mm7_be', 1, 0x113c),
        Register('mm7_bf', 1, 0x113d),
        Register('mm7_wd', 2, 0x113e),
        Register('mm7_bg', 1, 0x113e),
        Register('mm7_bh', 1, 0x113f),
        Register('ymm0', 32, 0x1200),
        Register('xmm0', 16, 0x1200),
        Register('xmm0_qa', 8, 0x1200),
        Register('xmm0_da', 4, 0x1200),
        Register('xmm0_wa', 2, 0x1200),
        Register('xmm0_ba', 1, 0x1200),
        Register('xmm0_bb', 1, 0x1201),
        Register('xmm0_wb', 2, 0x1202),
        Register('xmm0_bc', 1, 0x1202),
        Register('xmm0_bd', 1, 0x1203),
        Register('xmm0_db', 4, 0x1204),
        Register('xmm0_wc', 2, 0x1204),
        Register('xmm0_be', 1, 0x1204),
        Register('xmm0_bf', 1, 0x1205),
        Register('xmm0_wd', 2, 0x1206),
        Register('xmm0_bg', 1, 0x1206),
        Register('xmm0_bh', 1, 0x1207),
        Register('xmm0_qb', 8, 0x1208),
        Register('xmm0_dc', 4, 0x1208),
        Register('xmm0_we', 2, 0x1208),
        Register('xmm0_bi', 1, 0x1208),
        Register('xmm0_bj', 1, 0x1209),
        Register('xmm0_wf', 2, 0x120a),
        Register('xmm0_bk', 1, 0x120a),
        Register('xmm0_bl', 1, 0x120b),
        Register('xmm0_dd', 4, 0x120c),
        Register('xmm0_wg', 2, 0x120c),
        Register('xmm0_bm', 1, 0x120c),
        Register('xmm0_bn', 1, 0x120d),
        Register('xmm0_wh', 2, 0x120e),
        Register('xmm0_bo', 1, 0x120e),
        Register('xmm0_bp', 1, 0x120f),
        Register('ymm0_h', 16, 0x1210),
        Register('ymm1', 32, 0x1220),
        Register('xmm1', 16, 0x1220),
        Register('xmm1_qa', 8, 0x1220),
        Register('xmm1_da', 4, 0x1220),
        Register('xmm1_wa', 2, 0x1220),
        Register('xmm1_ba', 1, 0x1220),
        Register('xmm1_bb', 1, 0x1221),
        Register('xmm1_wb', 2, 0x1222),
        Register('xmm1_bc', 1, 0x1222),
        Register('xmm1_bd', 1, 0x1223),
        Register('xmm1_db', 4, 0x1224),
        Register('xmm1_wc', 2, 0x1224),
        Register('xmm1_be', 1, 0x1224),
        Register('xmm1_bf', 1, 0x1225),
        Register('xmm1_wd', 2, 0x1226),
        Register('xmm1_bg', 1, 0x1226),
        Register('xmm1_bh', 1, 0x1227),
        Register('xmm1_qb', 8, 0x1228),
        Register('xmm1_dc', 4, 0x1228),
        Register('xmm1_we', 2, 0x1228),
        Register('xmm1_bi', 1, 0x1228),
        Register('xmm1_bj', 1, 0x1229),
        Register('xmm1_wf', 2, 0x122a),
        Register('xmm1_bk', 1, 0x122a),
        Register('xmm1_bl', 1, 0x122b),
        Register('xmm1_dd', 4, 0x122c),
        Register('xmm1_wg', 2, 0x122c),
        Register('xmm1_bm', 1, 0x122c),
        Register('xmm1_bn', 1, 0x122d),
        Register('xmm1_wh', 2, 0x122e),
        Register('xmm1_bo', 1, 0x122e),
        Register('xmm1_bp', 1, 0x122f),
        Register('ymm1_h', 16, 0x1230),
        Register('ymm2', 32, 0x1240),
        Register('xmm2', 16, 0x1240),
        Register('xmm2_qa', 8, 0x1240),
        Register('xmm2_da', 4, 0x1240),
        Register('xmm2_wa', 2, 0x1240),
        Register('xmm2_ba', 1, 0x1240),
        Register('xmm2_bb', 1, 0x1241),
        Register('xmm2_wb', 2, 0x1242),
        Register('xmm2_bc', 1, 0x1242),
        Register('xmm2_bd', 1, 0x1243),
        Register('xmm2_db', 4, 0x1244),
        Register('xmm2_wc', 2, 0x1244),
        Register('xmm2_be', 1, 0x1244),
        Register('xmm2_bf', 1, 0x1245),
        Register('xmm2_wd', 2, 0x1246),
        Register('xmm2_bg', 1, 0x1246),
        Register('xmm2_bh', 1, 0x1247),
        Register('xmm2_qb', 8, 0x1248),
        Register('xmm2_dc', 4, 0x1248),
        Register('xmm2_we', 2, 0x1248),
        Register('xmm2_bi', 1, 0x1248),
        Register('xmm2_bj', 1, 0x1249),
        Register('xmm2_wf', 2, 0x124a),
        Register('xmm2_bk', 1, 0x124a),
        Register('xmm2_bl', 1, 0x124b),
        Register('xmm2_dd', 4, 0x124c),
        Register('xmm2_wg', 2, 0x124c),
        Register('xmm2_bm', 1, 0x124c),
        Register('xmm2_bn', 1, 0x124d),
        Register('xmm2_wh', 2, 0x124e),
        Register('xmm2_bo', 1, 0x124e),
        Register('xmm2_bp', 1, 0x124f),
        Register('ymm2_h', 16, 0x1250),
        Register('ymm3', 32, 0x1260),
        Register('xmm3', 16, 0x1260),
        Register('xmm3_qa', 8, 0x1260),
        Register('xmm3_da', 4, 0x1260),
        Register('xmm3_wa', 2, 0x1260),
        Register('xmm3_ba', 1, 0x1260),
        Register('xmm3_bb', 1, 0x1261),
        Register('xmm3_wb', 2, 0x1262),
        Register('xmm3_bc', 1, 0x1262),
        Register('xmm3_bd', 1, 0x1263),
        Register('xmm3_db', 4, 0x1264),
        Register('xmm3_wc', 2, 0x1264),
        Register('xmm3_be', 1, 0x1264),
        Register('xmm3_bf', 1, 0x1265),
        Register('xmm3_wd', 2, 0x1266),
        Register('xmm3_bg', 1, 0x1266),
        Register('xmm3_bh', 1, 0x1267),
        Register('xmm3_qb', 8, 0x1268),
        Register('xmm3_dc', 4, 0x1268),
        Register('xmm3_we', 2, 0x1268),
        Register('xmm3_bi', 1, 0x1268),
        Register('xmm3_bj', 1, 0x1269),
        Register('xmm3_wf', 2, 0x126a),
        Register('xmm3_bk', 1, 0x126a),
        Register('xmm3_bl', 1, 0x126b),
        Register('xmm3_dd', 4, 0x126c),
        Register('xmm3_wg', 2, 0x126c),
        Register('xmm3_bm', 1, 0x126c),
        Register('xmm3_bn', 1, 0x126d),
        Register('xmm3_wh', 2, 0x126e),
        Register('xmm3_bo', 1, 0x126e),
        Register('xmm3_bp', 1, 0x126f),
        Register('ymm3_h', 16, 0x1270),
        Register('ymm4', 32, 0x1280),
        Register('xmm4', 16, 0x1280),
        Register('xmm4_qa', 8, 0x1280),
        Register('xmm4_da', 4, 0x1280),
        Register('xmm4_wa', 2, 0x1280),
        Register('xmm4_ba', 1, 0x1280),
        Register('xmm4_bb', 1, 0x1281),
        Register('xmm4_wb', 2, 0x1282),
        Register('xmm4_bc', 1, 0x1282),
        Register('xmm4_bd', 1, 0x1283),
        Register('xmm4_db', 4, 0x1284),
        Register('xmm4_wc', 2, 0x1284),
        Register('xmm4_be', 1, 0x1284),
        Register('xmm4_bf', 1, 0x1285),
        Register('xmm4_wd', 2, 0x1286),
        Register('xmm4_bg', 1, 0x1286),
        Register('xmm4_bh', 1, 0x1287),
        Register('xmm4_qb', 8, 0x1288),
        Register('xmm4_dc', 4, 0x1288),
        Register('xmm4_we', 2, 0x1288),
        Register('xmm4_bi', 1, 0x1288),
        Register('xmm4_bj', 1, 0x1289),
        Register('xmm4_wf', 2, 0x128a),
        Register('xmm4_bk', 1, 0x128a),
        Register('xmm4_bl', 1, 0x128b),
        Register('xmm4_dd', 4, 0x128c),
        Register('xmm4_wg', 2, 0x128c),
        Register('xmm4_bm', 1, 0x128c),
        Register('xmm4_bn', 1, 0x128d),
        Register('xmm4_wh', 2, 0x128e),
        Register('xmm4_bo', 1, 0x128e),
        Register('xmm4_bp', 1, 0x128f),
        Register('ymm4_h', 16, 0x1290),
        Register('ymm5', 32, 0x12a0),
        Register('xmm5', 16, 0x12a0),
        Register('xmm5_qa', 8, 0x12a0),
        Register('xmm5_da', 4, 0x12a0),
        Register('xmm5_wa', 2, 0x12a0),
        Register('xmm5_ba', 1, 0x12a0),
        Register('xmm5_bb', 1, 0x12a1),
        Register('xmm5_wb', 2, 0x12a2),
        Register('xmm5_bc', 1, 0x12a2),
        Register('xmm5_bd', 1, 0x12a3),
        Register('xmm5_db', 4, 0x12a4),
        Register('xmm5_wc', 2, 0x12a4),
        Register('xmm5_be', 1, 0x12a4),
        Register('xmm5_bf', 1, 0x12a5),
        Register('xmm5_wd', 2, 0x12a6),
        Register('xmm5_bg', 1, 0x12a6),
        Register('xmm5_bh', 1, 0x12a7),
        Register('xmm5_qb', 8, 0x12a8),
        Register('xmm5_dc', 4, 0x12a8),
        Register('xmm5_we', 2, 0x12a8),
        Register('xmm5_bi', 1, 0x12a8),
        Register('xmm5_bj', 1, 0x12a9),
        Register('xmm5_wf', 2, 0x12aa),
        Register('xmm5_bk', 1, 0x12aa),
        Register('xmm5_bl', 1, 0x12ab),
        Register('xmm5_dd', 4, 0x12ac),
        Register('xmm5_wg', 2, 0x12ac),
        Register('xmm5_bm', 1, 0x12ac),
        Register('xmm5_bn', 1, 0x12ad),
        Register('xmm5_wh', 2, 0x12ae),
        Register('xmm5_bo', 1, 0x12ae),
        Register('xmm5_bp', 1, 0x12af),
        Register('ymm5_h', 16, 0x12b0),
        Register('ymm6', 32, 0x12c0),
        Register('xmm6', 16, 0x12c0),
        Register('xmm6_qa', 8, 0x12c0),
        Register('xmm6_da', 4, 0x12c0),
        Register('xmm6_wa', 2, 0x12c0),
        Register('xmm6_ba', 1, 0x12c0),
        Register('xmm6_bb', 1, 0x12c1),
        Register('xmm6_wb', 2, 0x12c2),
        Register('xmm6_bc', 1, 0x12c2),
        Register('xmm6_bd', 1, 0x12c3),
        Register('xmm6_db', 4, 0x12c4),
        Register('xmm6_wc', 2, 0x12c4),
        Register('xmm6_be', 1, 0x12c4),
        Register('xmm6_bf', 1, 0x12c5),
        Register('xmm6_wd', 2, 0x12c6),
        Register('xmm6_bg', 1, 0x12c6),
        Register('xmm6_bh', 1, 0x12c7),
        Register('xmm6_qb', 8, 0x12c8),
        Register('xmm6_dc', 4, 0x12c8),
        Register('xmm6_we', 2, 0x12c8),
        Register('xmm6_bi', 1, 0x12c8),
        Register('xmm6_bj', 1, 0x12c9),
        Register('xmm6_wf', 2, 0x12ca),
        Register('xmm6_bk', 1, 0x12ca),
        Register('xmm6_bl', 1, 0x12cb),
        Register('xmm6_dd', 4, 0x12cc),
        Register('xmm6_wg', 2, 0x12cc),
        Register('xmm6_bm', 1, 0x12cc),
        Register('xmm6_bn', 1, 0x12cd),
        Register('xmm6_wh', 2, 0x12ce),
        Register('xmm6_bo', 1, 0x12ce),
        Register('xmm6_bp', 1, 0x12cf),
        Register('ymm6_h', 16, 0x12d0),
        Register('ymm7', 32, 0x12e0),
        Register('xmm7', 16, 0x12e0),
        Register('xmm7_qa', 8, 0x12e0),
        Register('xmm7_da', 4, 0x12e0),
        Register('xmm7_wa', 2, 0x12e0),
        Register('xmm7_ba', 1, 0x12e0),
        Register('xmm7_bb', 1, 0x12e1),
        Register('xmm7_wb', 2, 0x12e2),
        Register('xmm7_bc', 1, 0x12e2),
        Register('xmm7_bd', 1, 0x12e3),
        Register('xmm7_db', 4, 0x12e4),
        Register('xmm7_wc', 2, 0x12e4),
        Register('xmm7_be', 1, 0x12e4),
        Register('xmm7_bf', 1, 0x12e5),
        Register('xmm7_wd', 2, 0x12e6),
        Register('xmm7_bg', 1, 0x12e6),
        Register('xmm7_bh', 1, 0x12e7),
        Register('xmm7_qb', 8, 0x12e8),
        Register('xmm7_dc', 4, 0x12e8),
        Register('xmm7_we', 2, 0x12e8),
        Register('xmm7_bi', 1, 0x12e8),
        Register('xmm7_bj', 1, 0x12e9),
        Register('xmm7_wf', 2, 0x12ea),
        Register('xmm7_bk', 1, 0x12ea),
        Register('xmm7_bl', 1, 0x12eb),
        Register('xmm7_dd', 4, 0x12ec),
        Register('xmm7_wg', 2, 0x12ec),
        Register('xmm7_bm', 1, 0x12ec),
        Register('xmm7_bn', 1, 0x12ed),
        Register('xmm7_wh', 2, 0x12ee),
        Register('xmm7_bo', 1, 0x12ee),
        Register('xmm7_bp', 1, 0x12ef),
        Register('ymm7_h', 16, 0x12f0),
        Register('ymm8', 32, 0x1300),
        Register('xmm8', 16, 0x1300),
        Register('xmm8_qa', 8, 0x1300),
        Register('xmm8_da', 4, 0x1300),
        Register('xmm8_wa', 2, 0x1300),
        Register('xmm8_ba', 1, 0x1300),
        Register('xmm8_bb', 1, 0x1301),
        Register('xmm8_wb', 2, 0x1302),
        Register('xmm8_bc', 1, 0x1302),
        Register('xmm8_bd', 1, 0x1303),
        Register('xmm8_db', 4, 0x1304),
        Register('xmm8_wc', 2, 0x1304),
        Register('xmm8_be', 1, 0x1304),
        Register('xmm8_bf', 1, 0x1305),
        Register('xmm8_wd', 2, 0x1306),
        Register('xmm8_bg', 1, 0x1306),
        Register('xmm8_bh', 1, 0x1307),
        Register('xmm8_qb', 8, 0x1308),
        Register('xmm8_dc', 4, 0x1308),
        Register('xmm8_we', 2, 0x1308),
        Register('xmm8_bi', 1, 0x1308),
        Register('xmm8_bj', 1, 0x1309),
        Register('xmm8_wf', 2, 0x130a),
        Register('xmm8_bk', 1, 0x130a),
        Register('xmm8_bl', 1, 0x130b),
        Register('xmm8_dd', 4, 0x130c),
        Register('xmm8_wg', 2, 0x130c),
        Register('xmm8_bm', 1, 0x130c),
        Register('xmm8_bn', 1, 0x130d),
        Register('xmm8_wh', 2, 0x130e),
        Register('xmm8_bo', 1, 0x130e),
        Register('xmm8_bp', 1, 0x130f),
        Register('ymm8_h', 16, 0x1310),
        Register('ymm9', 32, 0x1320),
        Register('xmm9', 16, 0x1320),
        Register('xmm9_qa', 8, 0x1320),
        Register('xmm9_da', 4, 0x1320),
        Register('xmm9_wa', 2, 0x1320),
        Register('xmm9_ba', 1, 0x1320),
        Register('xmm9_bb', 1, 0x1321),
        Register('xmm9_wb', 2, 0x1322),
        Register('xmm9_bc', 1, 0x1322),
        Register('xmm9_bd', 1, 0x1323),
        Register('xmm9_db', 4, 0x1324),
        Register('xmm9_wc', 2, 0x1324),
        Register('xmm9_be', 1, 0x1324),
        Register('xmm9_bf', 1, 0x1325),
        Register('xmm9_wd', 2, 0x1326),
        Register('xmm9_bg', 1, 0x1326),
        Register('xmm9_bh', 1, 0x1327),
        Register('xmm9_qb', 8, 0x1328),
        Register('xmm9_dc', 4, 0x1328),
        Register('xmm9_we', 2, 0x1328),
        Register('xmm9_bi', 1, 0x1328),
        Register('xmm9_bj', 1, 0x1329),
        Register('xmm9_wf', 2, 0x132a),
        Register('xmm9_bk', 1, 0x132a),
        Register('xmm9_bl', 1, 0x132b),
        Register('xmm9_dd', 4, 0x132c),
        Register('xmm9_wg', 2, 0x132c),
        Register('xmm9_bm', 1, 0x132c),
        Register('xmm9_bn', 1, 0x132d),
        Register('xmm9_wh', 2, 0x132e),
        Register('xmm9_bo', 1, 0x132e),
        Register('xmm9_bp', 1, 0x132f),
        Register('ymm9_h', 16, 0x1330),
        Register('ymm10', 32, 0x1340),
        Register('xmm10', 16, 0x1340),
        Register('xmm10_qa', 8, 0x1340),
        Register('xmm10_da', 4, 0x1340),
        Register('xmm10_wa', 2, 0x1340),
        Register('xmm10_ba', 1, 0x1340),
        Register('xmm10_bb', 1, 0x1341),
        Register('xmm10_wb', 2, 0x1342),
        Register('xmm10_bc', 1, 0x1342),
        Register('xmm10_bd', 1, 0x1343),
        Register('xmm10_db', 4, 0x1344),
        Register('xmm10_wc', 2, 0x1344),
        Register('xmm10_be', 1, 0x1344),
        Register('xmm10_bf', 1, 0x1345),
        Register('xmm10_wd', 2, 0x1346),
        Register('xmm10_bg', 1, 0x1346),
        Register('xmm10_bh', 1, 0x1347),
        Register('xmm10_qb', 8, 0x1348),
        Register('xmm10_dc', 4, 0x1348),
        Register('xmm10_we', 2, 0x1348),
        Register('xmm10_bi', 1, 0x1348),
        Register('xmm10_bj', 1, 0x1349),
        Register('xmm10_wf', 2, 0x134a),
        Register('xmm10_bk', 1, 0x134a),
        Register('xmm10_bl', 1, 0x134b),
        Register('xmm10_dd', 4, 0x134c),
        Register('xmm10_wg', 2, 0x134c),
        Register('xmm10_bm', 1, 0x134c),
        Register('xmm10_bn', 1, 0x134d),
        Register('xmm10_wh', 2, 0x134e),
        Register('xmm10_bo', 1, 0x134e),
        Register('xmm10_bp', 1, 0x134f),
        Register('ymm10_h', 16, 0x1350),
        Register('ymm11', 32, 0x1360),
        Register('xmm11', 16, 0x1360),
        Register('xmm11_qa', 8, 0x1360),
        Register('xmm11_da', 4, 0x1360),
        Register('xmm11_wa', 2, 0x1360),
        Register('xmm11_ba', 1, 0x1360),
        Register('xmm11_bb', 1, 0x1361),
        Register('xmm11_wb', 2, 0x1362),
        Register('xmm11_bc', 1, 0x1362),
        Register('xmm11_bd', 1, 0x1363),
        Register('xmm11_db', 4, 0x1364),
        Register('xmm11_wc', 2, 0x1364),
        Register('xmm11_be', 1, 0x1364),
        Register('xmm11_bf', 1, 0x1365),
        Register('xmm11_wd', 2, 0x1366),
        Register('xmm11_bg', 1, 0x1366),
        Register('xmm11_bh', 1, 0x1367),
        Register('xmm11_qb', 8, 0x1368),
        Register('xmm11_dc', 4, 0x1368),
        Register('xmm11_we', 2, 0x1368),
        Register('xmm11_bi', 1, 0x1368),
        Register('xmm11_bj', 1, 0x1369),
        Register('xmm11_wf', 2, 0x136a),
        Register('xmm11_bk', 1, 0x136a),
        Register('xmm11_bl', 1, 0x136b),
        Register('xmm11_dd', 4, 0x136c),
        Register('xmm11_wg', 2, 0x136c),
        Register('xmm11_bm', 1, 0x136c),
        Register('xmm11_bn', 1, 0x136d),
        Register('xmm11_wh', 2, 0x136e),
        Register('xmm11_bo', 1, 0x136e),
        Register('xmm11_bp', 1, 0x136f),
        Register('ymm11_h', 16, 0x1370),
        Register('ymm12', 32, 0x1380),
        Register('xmm12', 16, 0x1380),
        Register('xmm12_qa', 8, 0x1380),
        Register('xmm12_da', 4, 0x1380),
        Register('xmm12_wa', 2, 0x1380),
        Register('xmm12_ba', 1, 0x1380),
        Register('xmm12_bb', 1, 0x1381),
        Register('xmm12_wb', 2, 0x1382),
        Register('xmm12_bc', 1, 0x1382),
        Register('xmm12_bd', 1, 0x1383),
        Register('xmm12_db', 4, 0x1384),
        Register('xmm12_wc', 2, 0x1384),
        Register('xmm12_be', 1, 0x1384),
        Register('xmm12_bf', 1, 0x1385),
        Register('xmm12_wd', 2, 0x1386),
        Register('xmm12_bg', 1, 0x1386),
        Register('xmm12_bh', 1, 0x1387),
        Register('xmm12_qb', 8, 0x1388),
        Register('xmm12_dc', 4, 0x1388),
        Register('xmm12_we', 2, 0x1388),
        Register('xmm12_bi', 1, 0x1388),
        Register('xmm12_bj', 1, 0x1389),
        Register('xmm12_wf', 2, 0x138a),
        Register('xmm12_bk', 1, 0x138a),
        Register('xmm12_bl', 1, 0x138b),
        Register('xmm12_dd', 4, 0x138c),
        Register('xmm12_wg', 2, 0x138c),
        Register('xmm12_bm', 1, 0x138c),
        Register('xmm12_bn', 1, 0x138d),
        Register('xmm12_wh', 2, 0x138e),
        Register('xmm12_bo', 1, 0x138e),
        Register('xmm12_bp', 1, 0x138f),
        Register('ymm12_h', 16, 0x1390),
        Register('ymm13', 32, 0x13a0),
        Register('xmm13', 16, 0x13a0),
        Register('xmm13_qa', 8, 0x13a0),
        Register('xmm13_da', 4, 0x13a0),
        Register('xmm13_wa', 2, 0x13a0),
        Register('xmm13_ba', 1, 0x13a0),
        Register('xmm13_bb', 1, 0x13a1),
        Register('xmm13_wb', 2, 0x13a2),
        Register('xmm13_bc', 1, 0x13a2),
        Register('xmm13_bd', 1, 0x13a3),
        Register('xmm13_db', 4, 0x13a4),
        Register('xmm13_wc', 2, 0x13a4),
        Register('xmm13_be', 1, 0x13a4),
        Register('xmm13_bf', 1, 0x13a5),
        Register('xmm13_wd', 2, 0x13a6),
        Register('xmm13_bg', 1, 0x13a6),
        Register('xmm13_bh', 1, 0x13a7),
        Register('xmm13_qb', 8, 0x13a8),
        Register('xmm13_dc', 4, 0x13a8),
        Register('xmm13_we', 2, 0x13a8),
        Register('xmm13_bi', 1, 0x13a8),
        Register('xmm13_bj', 1, 0x13a9),
        Register('xmm13_wf', 2, 0x13aa),
        Register('xmm13_bk', 1, 0x13aa),
        Register('xmm13_bl', 1, 0x13ab),
        Register('xmm13_dd', 4, 0x13ac),
        Register('xmm13_wg', 2, 0x13ac),
        Register('xmm13_bm', 1, 0x13ac),
        Register('xmm13_bn', 1, 0x13ad),
        Register('xmm13_wh', 2, 0x13ae),
        Register('xmm13_bo', 1, 0x13ae),
        Register('xmm13_bp', 1, 0x13af),
        Register('ymm13_h', 16, 0x13b0),
        Register('ymm14', 32, 0x13c0),
        Register('xmm14', 16, 0x13c0),
        Register('xmm14_qa', 8, 0x13c0),
        Register('xmm14_da', 4, 0x13c0),
        Register('xmm14_wa', 2, 0x13c0),
        Register('xmm14_ba', 1, 0x13c0),
        Register('xmm14_bb', 1, 0x13c1),
        Register('xmm14_wb', 2, 0x13c2),
        Register('xmm14_bc', 1, 0x13c2),
        Register('xmm14_bd', 1, 0x13c3),
        Register('xmm14_db', 4, 0x13c4),
        Register('xmm14_wc', 2, 0x13c4),
        Register('xmm14_be', 1, 0x13c4),
        Register('xmm14_bf', 1, 0x13c5),
        Register('xmm14_wd', 2, 0x13c6),
        Register('xmm14_bg', 1, 0x13c6),
        Register('xmm14_bh', 1, 0x13c7),
        Register('xmm14_qb', 8, 0x13c8),
        Register('xmm14_dc', 4, 0x13c8),
        Register('xmm14_we', 2, 0x13c8),
        Register('xmm14_bi', 1, 0x13c8),
        Register('xmm14_bj', 1, 0x13c9),
        Register('xmm14_wf', 2, 0x13ca),
        Register('xmm14_bk', 1, 0x13ca),
        Register('xmm14_bl', 1, 0x13cb),
        Register('xmm14_dd', 4, 0x13cc),
        Register('xmm14_wg', 2, 0x13cc),
        Register('xmm14_bm', 1, 0x13cc),
        Register('xmm14_bn', 1, 0x13cd),
        Register('xmm14_wh', 2, 0x13ce),
        Register('xmm14_bo', 1, 0x13ce),
        Register('xmm14_bp', 1, 0x13cf),
        Register('ymm14_h', 16, 0x13d0),
        Register('ymm15', 32, 0x13e0),
        Register('xmm15', 16, 0x13e0),
        Register('xmm15_qa', 8, 0x13e0),
        Register('xmm15_da', 4, 0x13e0),
        Register('xmm15_wa', 2, 0x13e0),
        Register('xmm15_ba', 1, 0x13e0),
        Register('xmm15_bb', 1, 0x13e1),
        Register('xmm15_wb', 2, 0x13e2),
        Register('xmm15_bc', 1, 0x13e2),
        Register('xmm15_bd', 1, 0x13e3),
        Register('xmm15_db', 4, 0x13e4),
        Register('xmm15_wc', 2, 0x13e4),
        Register('xmm15_be', 1, 0x13e4),
        Register('xmm15_bf', 1, 0x13e5),
        Register('xmm15_wd', 2, 0x13e6),
        Register('xmm15_bg', 1, 0x13e6),
        Register('xmm15_bh', 1, 0x13e7),
        Register('xmm15_qb', 8, 0x13e8),
        Register('xmm15_dc', 4, 0x13e8),
        Register('xmm15_we', 2, 0x13e8),
        Register('xmm15_bi', 1, 0x13e8),
        Register('xmm15_bj', 1, 0x13e9),
        Register('xmm15_wf', 2, 0x13ea),
        Register('xmm15_bk', 1, 0x13ea),
        Register('xmm15_bl', 1, 0x13eb),
        Register('xmm15_dd', 4, 0x13ec),
        Register('xmm15_wg', 2, 0x13ec),
        Register('xmm15_bm', 1, 0x13ec),
        Register('xmm15_bn', 1, 0x13ed),
        Register('xmm15_wh', 2, 0x13ee),
        Register('xmm15_bo', 1, 0x13ee),
        Register('xmm15_bp', 1, 0x13ef),
        Register('ymm15_h', 16, 0x13f0),
        Register('xmmtmp1', 16, 0x1400),
        Register('xmmtmp1_qa', 8, 0x1400),
        Register('xmmtmp1_da', 4, 0x1400),
        Register('xmmtmp1_db', 4, 0x1404),
        Register('xmmtmp1_qb', 8, 0x1408),
        Register('xmmtmp1_dc', 4, 0x1408),
        Register('xmmtmp1_dd', 4, 0x140c),
        Register('xmmtmp2', 16, 0x1410),
        Register('xmmtmp2_qa', 8, 0x1410),
        Register('xmmtmp2_da', 4, 0x1410),
        Register('xmmtmp2_db', 4, 0x1414),
        Register('xmmtmp2_qb', 8, 0x1418),
        Register('xmmtmp2_dc', 4, 0x1418),
        Register('xmmtmp2_dd', 4, 0x141c),
        Register('contextreg', 4, 0x2000),
        Register('idtr', 6, 0x2200),
        Register('idtr_limit', 2, 0x2200),
        Register('idtr_address', 4, 0x2202),
        Register('gdtr', 6, 0x2210),
        Register('gdtr_limit', 2, 0x2210),
        Register('gdtr_address', 4, 0x2212),
        Register('ldtr', 6, 0x2220),
        Register('ldtr_limit', 2, 0x2220),
        Register('ldtr_address', 4, 0x2222),
        Register('tr', 6, 0x2230),
        Register('tr_limit', 2, 0x2230),
        Register('tr_address', 4, 0x2232)
    ]

register_arch(['x86:le:32:system management mode'], 32, Endness.LE, ArchPcode_x86_LE_32_System_Management_Mode)



from typing import Optional, Set, List, Tuple, TYPE_CHECKING
import logging

import claripy

from ...storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
from ...engines.light import SimEngineLight, ArithmeticExpression
from ...errors import SimEngineError, SimMemoryMissingError
from ...sim_variable import SimVariable, SimStackVariable, SimRegisterVariable, SimMemoryVariable
from ...code_location import CodeLocation
from ..typehoon import typevars, typeconsts

if TYPE_CHECKING:
    from .variable_recovery_base import VariableRecoveryStateBase
    from angr.knowledge_plugins.variables.variable_manager import VariableManager

#
# The base engine used in VariableRecoveryFast
#

l = logging.getLogger(name=__name__)


class RichR:
    """
    A rich representation of calculation results. The variable recovery data domain.
    """

    __slots__ = ('data', 'variable', 'typevar', 'type_constraints', )

    def __init__(self, data: claripy.ast.Base, variable=None, typevar: Optional[typevars.TypeVariable]=None,
                 type_constraints=None):
        self.data: claripy.ast.Base = data
        self.variable = variable
        self.typevar = typevar
        self.type_constraints = type_constraints

    @property
    def bits(self):
        if self.data is not None and not isinstance(self.data, (int, float)):
            if isinstance(self.data, claripy.ast.Base):
                return self.data.size()
            return self.data.bits
        if self.variable is not None:
            return self.variable.bits
        return None

    def __repr__(self):
        return "R{%r}" % self.data


class SimEngineVRBase(SimEngineLight):
    """
    The base class for variable recovery analyses. Contains methods for basic interactions with the state, like loading
    and storing data.
    """

    state: 'VariableRecoveryStateBase'

    def __init__(self, project, kb):
        super().__init__()

        self.project = project
        self.kb = kb
        self.variable_manager: Optional['VariableManager'] = None

    @property
    def func_addr(self):
        if self.state is None:
            return None
        return self.state.function.addr

    def process(self, state, *args, **kwargs):  # pylint:disable=unused-argument

        self.variable_manager = state.variable_manager

        try:
            self._process(state, None, block=kwargs.pop('block', None))
        except SimEngineError as e:
            if kwargs.pop('fail_fast', False) is True:
                raise e

    def _process(self, state, successors, block=None, func_addr=None):  # pylint:disable=unused-argument,arguments-differ
        super()._process(state, successors, block=block)

    #
    # Address parsing
    #

    @staticmethod
    def _addr_has_concrete_base(addr: claripy.ast.BV) -> bool:
        if addr.op == "__add__":
            if len(addr.args) == 2:
                if addr.args[0].concrete:
                    return True
                if addr.args[1].concrete:
                    return True
        return False

    @staticmethod
    def _parse_offseted_addr(addr: claripy.ast.BV) -> Optional[Tuple[claripy.ast.BV,claripy.ast.BV,claripy.ast.BV]]:
        if addr.op == "__add__":
            if len(addr.args) == 2:
                concrete_base, byte_offset = None, None
                if addr.args[0].concrete:
                    concrete_base, byte_offset = addr.args
                elif addr.args[1].concrete:
                    concrete_base, byte_offset = addr.args[1], addr.args[0]
                if concrete_base is None or byte_offset is None:
                    return None
                base_addr = concrete_base
                offset = None
                elem_size = None
                if byte_offset.concrete:
                    offset = byte_offset
                    elem_size = 1
                else:
                    abs_offset = byte_offset
                    if abs_offset.op == "__lshift__" and abs_offset.args[1].concrete:
                        offset = abs_offset.args[0]
                        elem_size = 2 ** abs_offset.args[1]._model_concrete.value

                if base_addr is not None and offset is not None and elem_size is not None:
                    return base_addr, offset, elem_size
        return None

    #
    # Logic
    #

    def _reference(self, richr: RichR, codeloc: CodeLocation, src=None):
        data: claripy.ast.Base = richr.data

        if data is None:
            return

        if self.state.is_stack_address(data):
            # this is a stack address
            # extract stack offset
            stack_offset: Optional[int] = self.state.get_stack_offset(data)

            variable_manager = self.variable_manager[self.func_addr]
            var_candidates: List[Tuple[SimVariable, int]] = variable_manager.find_variables_by_stmt(
                self.block.addr,
                self.stmt_idx,
                'memory')

            # find the correct variable
            existing_vars: List[Tuple[SimVariable, int]] = []
            for candidate, offset in var_candidates:
                if isinstance(candidate, SimStackVariable) and candidate.offset == stack_offset:
                    existing_vars.append((candidate, offset))
            variable = None
            if existing_vars:
                variable, _ = existing_vars[0]

            vs = None
            if stack_offset is not None:
                stack_addr = self.state.stack_addr_from_offset(stack_offset)
                if variable is None:
                    # TODO: how to determine the size for a lea?
                    try:
                        vs: Optional[MultiValues] = self.state.stack_region.load(stack_addr, size=1)
                    except SimMemoryMissingError:
                        vs = None

                    if vs is not None:
                        # extract variables
                        for values in vs.values.values():
                            for v in values:
                                for var_stack_offset, var in self.state.extract_variables(v):
                                    existing_vars.append((var, var_stack_offset))

                    if not existing_vars:
                        # no variables exist
                        lea_size = 1
                        variable = SimStackVariable(stack_offset, lea_size, base='bp',
                                                    ident=self.variable_manager[self.func_addr].next_variable_ident(
                                                        'stack'),
                                                    region=self.func_addr,
                                                    )
                        self.variable_manager[self.func_addr].add_variable('stack', stack_offset, variable)
                        l.debug('Identified a new stack variable %s at %#x.', variable, self.ins_addr)
                        existing_vars.append((variable, 0))

                    else:
                        # FIXME: Why is it only taking the first variable?
                        variable = next(iter(existing_vars))[0]

                # write the variable back to stack
                if vs is None:
                    top = self.state.top(self.arch.byte_width)
                    top = self.state.annotate_with_variables(top, [(0, variable)])
                    vs = MultiValues(offset_to_values={0: {top}})
                self.state.stack_region.store(stack_addr, vs)

        elif self.state.is_global_variable_address(data):
            # this is probably an address for a global variable
            global_var_addr = data._model_concrete.value

            variable_manager = self.variable_manager['global']

            # special case for global variables: find existing variable by base address
            existing_vars = list((var, 0) for var in variable_manager.get_global_variables(global_var_addr))

            if not existing_vars:
                variable = SimMemoryVariable(global_var_addr, 1,
                                             ident=variable_manager.next_variable_ident('global'),
                                             )
                variable_manager.set_variable('global', global_var_addr, variable)
                l.debug('Identified a new global variable %s at %#x.', variable, self.ins_addr)
                existing_vars = [ (variable, 0) ]
            else:
                variable, _ = next(iter(existing_vars))

        else:
            return

        if self.state.typevars.has_type_variable_for(variable, codeloc):
            variable_typevar = self.state.typevars.get_type_variable(variable, codeloc)
        else:
            variable_typevar = typevars.TypeVariable()
            self.state.typevars.add_type_variable(variable, codeloc, variable_typevar)
        addr_typevar = typevars.TypeVariable() if richr.typevar is None else richr.typevar
        derived_typevar = typevars.DerivedTypeVariable(
            typevars.DerivedTypeVariable(addr_typevar, typevars.Load()),
            typevars.HasField(1 * 8, 0)  # at least one byte
        )
        type_constraint = typevars.Subtype(variable_typevar, derived_typevar)
        self.state.add_type_constraint(type_constraint)

        # find all variables
        for var, offset in existing_vars:
            if offset == 0:
                offset = None
            variable_manager.reference_at(var, offset, codeloc, atom=src)

    def _assign_to_register(self, offset, richr, size, src=None, dst=None):
        """

        :param int offset:
        :param RichR data:
        :param int size:
        :return:
        """

        if self.arch.is_artificial_register(offset, size):
            # skip all writes to artificial registers
            return
        if offset in (self.arch.ip_offset, self.arch.sp_offset):
            # only store the value. don't worry about variables.
            v = MultiValues(offset_to_values={0: {richr.data}})
            self.state.register_region.store(offset, v)
            return

        codeloc: CodeLocation = self._codeloc()
        data: claripy.ast.Base = richr.data

        # lea
        self._reference(richr, codeloc, src=src)

        # handle register writes
        existing_vars = self.variable_manager[self.func_addr].find_variables_by_atom(self.block.addr, self.stmt_idx,
                                                                                     dst)
        existing_vars: Set[Tuple[SimVariable,int]]
        if not existing_vars:
            variable = SimRegisterVariable(offset, size,
                                           ident=self.variable_manager[self.func_addr].next_variable_ident(
                                               'register'),
                                           region=self.func_addr
                                           )
            self.variable_manager[self.func_addr].set_variable('register', offset, variable)
        else:
            variable, _ = next(iter(existing_vars))

        # FIXME: The offset does not have to be 0
        annotated_data = self.state.annotate_with_variables(data, [(0, variable)])
        v = MultiValues(offset_to_values={0: {annotated_data}})
        self.state.register_region.store(offset, v)
        # register with the variable manager
        self.variable_manager[self.func_addr].write_to(variable, None, codeloc, atom=dst)

        if richr.typevar is not None:
            if not self.state.typevars.has_type_variable_for(variable, codeloc):
                # assign a new type variable to it
                typevar = typevars.TypeVariable()
                self.state.typevars.add_type_variable(variable, codeloc, typevar)
                # create constraints
            else:
                typevar = self.state.typevars.get_type_variable(variable, codeloc)
            self.state.add_type_constraint(typevars.Subtype(richr.typevar, typevar))
            self.state.add_type_constraint(typevars.Subtype(typevar, typeconsts.int_type(variable.size * 8)))

    def _store(self, richr_addr: RichR, data: RichR, size, stmt=None):  # pylint:disable=unused-argument
        """

        :param RichR addr:
        :param RichR data:
        :param int size:
        :return:
        """

        addr: claripy.ast.Base = richr_addr.data
        stored = False

        if addr.concrete:
            # fully concrete. this is a global address
            self._store_to_global(addr._model_concrete.value, data, size, stmt=stmt)
            stored = True
        elif self._addr_has_concrete_base(addr) and self._parse_offseted_addr(addr) is not None:
            # we are storing to a concrete global address with an offset
            base_addr, offset, elem_size = self._parse_offseted_addr(addr)
            self._store_to_global(base_addr._model_concrete.value, data, size,
                                  stmt=stmt, offset=offset, elem_size=elem_size)
            stored = True
        else:
            if self.state.is_stack_address(addr):
                stack_offset = self.state.get_stack_offset(addr)
                if stack_offset is not None:
                    # Storing data to stack
                    self._store_to_stack(stack_offset, data, size, stmt=stmt)
                    stored = True

        if not stored:
            # storing to a location specified by a pointer whose value cannot be determined at this point
            self._store_to_variable(richr_addr, size, stmt=stmt)

    def _store_to_stack(self, stack_offset, data: RichR, size, stmt=None, endness=None):
        if stmt is None:
            existing_vars = self.variable_manager[self.func_addr].find_variables_by_stmt(self.block.addr,
                                                                                         self.stmt_idx,
                                                                                         'memory'
                                                                                         )
        else:
            existing_vars = self.variable_manager[self.func_addr].find_variables_by_atom(self.block.addr,
                                                                                         self.stmt_idx,
                                                                                         stmt
                                                                                         )
        if not existing_vars:
            variable = SimStackVariable(stack_offset, size, base='bp',
                                        ident=self.variable_manager[self.func_addr].next_variable_ident(
                                            'stack'),
                                        region=self.func_addr,
                                        )
            variable_offset = 0
            if isinstance(stack_offset, int):
                self.variable_manager[self.func_addr].set_variable('stack', stack_offset, variable)
                l.debug('Identified a new stack variable %s at %#x.', variable, self.ins_addr)

        else:
            variable, variable_offset = next(iter(existing_vars))

        if isinstance(stack_offset, int):
            expr = self.state.annotate_with_variables(data.data, [(variable_offset, variable)])
            stack_addr = self.state.stack_addr_from_offset(stack_offset)
            self.state.stack_region.store(stack_addr, expr, endness=endness)

            codeloc = CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr)

            addr_and_variables = set()
            try:
                vs: MultiValues = self.state.stack_region.load(stack_addr, size, endness=endness)
                for values in vs.values.values():
                    for value in values:
                        addr_and_variables.update(self.state.extract_variables(value))
            except SimMemoryMissingError:
                pass

            for var_offset, var in addr_and_variables:
                offset_into_var = var_offset
                if offset_into_var == 0:
                    offset_into_var = None
                self.variable_manager[self.func_addr].write_to(var,
                                                               offset_into_var,
                                                               codeloc,
                                                               atom=stmt,
                                                               )

            # create type constraints
            if data.typevar is not None:
                if not self.state.typevars.has_type_variable_for(variable, codeloc):
                    typevar = typevars.TypeVariable()
                    self.state.typevars.add_type_variable(variable, codeloc, typevar)
                else:
                    typevar = self.state.typevars.get_type_variable(variable, codeloc)
                if typevar is not None:
                    self.state.add_type_constraint(
                        typevars.Subtype(data.typevar, typevar)
                    )
        # TODO: Create a tv_sp.store.<bits>@N <: typevar type constraint for the stack pointer

    def _store_to_global(self, addr: int, data: RichR, size: int, stmt=None, offset: Optional[claripy.ast.BV]=None,
                         elem_size: Optional[claripy.ast.BV]=None):
        variable_manager = self.variable_manager['global']
        if stmt is None:
            existing_vars = variable_manager.find_variables_by_stmt(self.block.addr, self.stmt_idx, 'memory')
        else:
            existing_vars = variable_manager.find_variables_by_atom(self.block.addr, self.stmt_idx, stmt)

        if offset is None or elem_size is None:
            # trivial case
            abs_addr = addr
        elif offset.concrete and elem_size.concrete:
            abs_addr = addr + offset._model_concrete.value * elem_size._model_concrete.value
        else:
            abs_addr = None

        if not existing_vars:
            # special case for global variables: find existing variable by base address
            existing_vars = { (var, (offset, elem_size)) for var in variable_manager.get_global_variables(addr) }

        if not existing_vars:
            variable = SimMemoryVariable(addr, size,
                                         ident=variable_manager.next_variable_ident('global'),
                                         )
            variable_manager.set_variable('global', addr, variable)
            l.debug('Identified a new global variable %s at %#x.', variable, self.ins_addr)
            existing_vars = {(variable, (offset, elem_size))}
        else:
            variable, _ = next(iter(existing_vars))

        data_expr: claripy.ast.Base = data.data
        data_expr = self.state.annotate_with_variables(data_expr, [(0, variable)])

        if abs_addr is not None:
            self.state.global_region.store(addr,
                                           data_expr,
                                           endness=self.state.arch.memory_endness if stmt is None else stmt.endness)

        codeloc = CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr)
        values = None
        if abs_addr is not None:
            try:
                values: MultiValues = self.state.global_region.load(
                    abs_addr,
                    size=size,
                    endness=self.state.arch.memory_endness if stmt is None else stmt.endness)
            except SimMemoryMissingError:
                pass

        if values is not None:
            for vs in values.values.values():
                for v in vs:
                    for var_offset, var in self.state.extract_variables(v):
                        variable_manager.write_to(var, var_offset, codeloc, atom=stmt)
        else:
            for var, var_offset in existing_vars:
                variable_manager.write_to(var, var_offset, codeloc, atom=stmt)

        # create type constraints
        if not self.state.typevars.has_type_variable_for(variable, codeloc):
            typevar = typevars.TypeVariable()
            self.state.typevars.add_type_variable(variable, codeloc, typevar)
        else:
            typevar = self.state.typevars.get_type_variable(variable, codeloc)

        if data.typevar is not None:
            self.state.add_type_constraint(
                typevars.Subtype(data.typevar, typevar)
            )

        if offset is not None and elem_size is not None:
            # it's an array!
            if offset.concrete and elem_size.concrete:
                concrete_offset = offset._model_concrete.value * elem_size._model_concrete.value
                store_typevar = typevars.DerivedTypeVariable(
                    typevars.DerivedTypeVariable(typevar, typevars.Store()),
                    typevars.HasField(size * self.state.arch.byte_width, concrete_offset)
                )
                self.state.add_type_constraint(
                    typevars.Existence(store_typevar)
                )
            else:
                # FIXME: This is a hack
                for i in range(0, 4):
                    concrete_offset = size * i
                    store_typevar = typevars.DerivedTypeVariable(
                        typevars.DerivedTypeVariable(typevar, typevars.Store()),
                        typevars.HasField(size * self.state.arch.byte_width, concrete_offset)
                    )
                    self.state.add_type_constraint(
                        typevars.Existence(store_typevar)
                    )

    def _store_to_variable(self, richr_addr: RichR, size: int, stmt=None):  # pylint:disable=unused-argument

        addr_variable = richr_addr.variable
        codeloc = self._codeloc()

        # Storing data into a pointer
        if richr_addr.type_constraints:
            for tc in richr_addr.type_constraints:
                self.state.add_type_constraint(tc)

        if richr_addr.typevar is None:
            typevar = typevars.TypeVariable()
        else:
            typevar = richr_addr.typevar

        if typevar is not None:
            if isinstance(typevar, typevars.DerivedTypeVariable) and isinstance(typevar.label, typevars.AddN):
                base_typevar = typevar.type_var
                field_offset = typevar.label.n
            else:
                base_typevar = typevar
                field_offset = 0

            # if addr_variable is not None:
            #     self.variable_manager[self.func_addr].reference_at(addr_variable, field_offset, codeloc, atom=stmt)

            store_typevar = typevars.DerivedTypeVariable(
                typevars.DerivedTypeVariable(base_typevar, typevars.Store()),
                typevars.HasField(size * self.state.arch.byte_width, field_offset)
            )
            if addr_variable is not None:
                self.state.typevars.add_type_variable(addr_variable, codeloc, typevar)
            self.state.add_type_constraint(typevars.Existence(store_typevar))

    def _load(self, richr_addr: RichR, size: int, expr=None):
        """

        :param RichR richr_addr:
        :param size:
        :return:
        """

        addr: claripy.ast.Base = richr_addr.data
        codeloc = CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr)
        typevar = None

        if self.state.is_stack_address(addr):
            stack_offset = self.state.get_stack_offset(addr)
            if stack_offset is not None:
                # Loading data from stack

                # split the offset into a concrete offset and a dynamic offset
                # the stack offset may not be a concrete offset
                # for example, SP-0xe0+var_1
                if type(stack_offset) is ArithmeticExpression:
                    if type(stack_offset.operands[0]) is int:
                        concrete_offset = stack_offset.operands[0]
                        dynamic_offset = stack_offset.operands[1]
                    elif type(stack_offset.operands[1]) is int:
                        concrete_offset = stack_offset.operands[1]
                        dynamic_offset = stack_offset.operands[0]
                    else:
                        # cannot determine the concrete offset. give up
                        concrete_offset = None
                        dynamic_offset = stack_offset
                else:
                    # type(stack_offset) is int
                    concrete_offset = stack_offset
                    dynamic_offset = None

                try:
                    values: Optional[MultiValues] = self.state.stack_region.load(
                        self.state.stack_addr_from_offset(concrete_offset),
                        size=size,
                        endness=self.state.arch.memory_endness)

                except SimMemoryMissingError:
                    values = None

                all_vars: Set[Tuple[int,SimVariable]] = set()
                if values:
                    for vs in values.values.values():
                        for v in vs:
                            for var_offset, var_ in self.state.extract_variables(v):
                                all_vars.add((var_offset, var_))

                if not all_vars:
                    variables = self.variable_manager[self.func_addr].find_variables_by_stack_offset(concrete_offset)
                    if not variables:
                        variable = SimStackVariable(concrete_offset, size, base='bp',
                                                    ident=self.variable_manager[self.func_addr].next_variable_ident(
                                                        'stack'),
                                                    region=self.func_addr,
                                                    )
                        self.variable_manager[self.func_addr].add_variable('stack', concrete_offset, variable)
                        variables = {variable}
                        l.debug('Identified a new stack variable %s at %#x.', variable, self.ins_addr)
                    for variable in variables:
                        v = self.state.top(size * self.state.arch.byte_width)
                        v = self.state.annotate_with_variables(v, [(0, variable)])
                        stack_addr = self.state.stack_addr_from_offset(concrete_offset)
                        self.state.stack_region.store(stack_addr, v, endness=self.state.arch.memory_endness)

                    all_vars = {(0, variable) for variable in variables}

                if len(all_vars) > 1:
                    # overlapping variables
                    l.warning("Reading memory with overlapping variables: %s. Ignoring all but the first one.",
                              all_vars)

                var_offset, var = next(iter(all_vars))  # won't fail
                # calculate variable_offset
                if dynamic_offset is None:
                    offset_into_variable = None
                else:
                    if var_offset == 0:
                        offset_into_variable = dynamic_offset
                    else:
                        offset_into_variable = ArithmeticExpression(ArithmeticExpression.Add,
                                                                    (dynamic_offset, var_offset,)
                                                                    )
                self.variable_manager[self.func_addr].read_from(var,
                                                                offset_into_variable,
                                                                codeloc,
                                                                atom=expr,
                                                                # overwrite=True
                                                                )

                # add delayed type constraints
                if var in self.state.delayed_type_constraints:
                    for constraint in self.state.delayed_type_constraints[var]:
                        self.state.add_type_constraint(constraint)
                    self.state.delayed_type_constraints.pop(var)
                # create type constraints
                if not self.state.typevars.has_type_variable_for(var, codeloc):
                    typevar = typevars.TypeVariable()
                    self.state.typevars.add_type_variable(var, codeloc, typevar)
                else:
                    typevar = self.state.typevars.get_type_variable(var, codeloc)
                # TODO: Create a tv_sp.load.<bits>@N type variable for the stack variable
                #typevar = typevars.DerivedTypeVariable(
                #    typevars.DerivedTypeVariable(typevar, typevars.Load()),
                #    typevars.HasField(size * 8, 0)
                #)

                r = self.state.top(size * self.state.arch.byte_width)
                r = self.state.annotate_with_variables(r, list(all_vars))
                return RichR(r, variable=var, typevar=typevar)

        elif addr.concrete:
            # Loading data from memory
            self._load_from_global(addr._model_concrete.value, size, expr=expr)

        elif self._addr_has_concrete_base(addr) and self._parse_offseted_addr(addr) is not None:
            # Loading data from a memory address with an offset
            base_addr, offset, elem_size = self._parse_offseted_addr(addr)
            self._load_from_global(base_addr._model_concrete.value, size, expr=expr, offset=offset,
                                   elem_size=elem_size)

        else:
            # Loading data from a pointer
            if richr_addr.type_constraints:
                for tc in richr_addr.type_constraints:
                    self.state.add_type_constraint(tc)

            # parse the loading offset
            offset = 0
            if (isinstance(richr_addr.typevar, typevars.DerivedTypeVariable) and
                    isinstance(richr_addr.typevar.label, typevars.AddN)):
                offset = richr_addr.typevar.label.n
                richr_addr_typevar = richr_addr.typevar.type_var  # unpack
            else:
                richr_addr_typevar = richr_addr.typevar

            if richr_addr_typevar is not None:
                # create a type constraint
                typevar = typevars.DerivedTypeVariable(
                    typevars.DerivedTypeVariable(richr_addr_typevar, typevars.Load()),
                    typevars.HasField(size * self.state.arch.byte_width, offset)
                )
                self.state.add_type_constraint(typevars.Existence(typevar))

        return RichR(self.state.top(size * self.state.arch.byte_width), typevar=typevar)

    def _load_from_global(self, addr: int, size, expr=None, offset: Optional[claripy.ast.BV]=None,
                          elem_size: Optional[claripy.ast.BV]=None):

        variable_manager = self.variable_manager['global']
        if expr is None:
            existing_vars = variable_manager.find_variables_by_stmt(self.block.addr, self.stmt_idx, 'memory')
        else:
            existing_vars = variable_manager.find_variables_by_atom(self.block.addr, self.stmt_idx, expr)

        # if offset is None or elem_size is None:
        #     # trivial case
        #     abs_addr = addr
        # elif offset.concrete and elem_size.concrete:
        #     abs_addr = addr + offset._model_concrete.value * elem_size._model_concrete.value
        # else:
        #     abs_addr = None

        if not existing_vars:
            # special case for global variables: find existing variable by base address
            existing_vars = { (var, (offset, elem_size)) for var in variable_manager.get_global_variables(addr) }

        if not existing_vars:
            # is this address mapped?
            if self.project.loader.find_object_containing(addr) is None:
                return
            variable = SimMemoryVariable(addr, size,
                                         ident=variable_manager.next_variable_ident('global'),
                                         )
            variable_manager.add_variable('global', addr, variable)
            l.debug('Identified a new global variable %s at %#x.', variable, self.ins_addr)
            existing_vars = {(variable, (offset, elem_size))}

        codeloc = CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr)
        for variable, _ in existing_vars:
            variable_manager.read_from(variable, None, codeloc, atom=expr)

        variable, _ = next(iter(existing_vars))
        # create type constraints
        if not self.state.typevars.has_type_variable_for(variable, codeloc):
            typevar = typevars.TypeVariable()
            self.state.typevars.add_type_variable(variable, codeloc, typevar)
        else:
            typevar = self.state.typevars.get_type_variable(variable, codeloc)

        if offset is not None and elem_size is not None:
            # it's an array!
            if offset.concrete and elem_size.concrete:
                concrete_offset = offset._model_concrete.value * elem_size._model_concrete.value
                load_typevar = typevars.DerivedTypeVariable(
                    typevars.DerivedTypeVariable(typevar, typevars.Store()),
                    typevars.HasField(size * self.state.arch.byte_width, concrete_offset)
                )
                self.state.add_type_constraint(
                    typevars.Existence(load_typevar)
                )
            else:
                # FIXME: This is a hack
                for i in range(0, 4):
                    concrete_offset = size * i
                    load_typevar = typevars.DerivedTypeVariable(
                        typevars.DerivedTypeVariable(typevar, typevars.Store()),
                        typevars.HasField(size * self.state.arch.byte_width, concrete_offset)
                    )
                    self.state.add_type_constraint(
                        typevars.Existence(load_typevar)
                    )

    def _read_from_register(self, offset, size, expr=None):
        """

        :param offset:
        :param size:
        :return:
        """

        codeloc = self._codeloc()

        try:
            values: Optional[MultiValues] = self.state.register_region.load(offset, size=size)
        except SimMemoryMissingError:
            values = None

        if self.arch.is_artificial_register(offset, size):
            # don't even load values for artificial registers
            r_value = self.state.top(size * self.arch.byte_width)
            return RichR(r_value, variable=None, typevar=None)
        if offset in (self.arch.sp_offset, self.arch.ip_offset):
            # load values. don't worry about variables
            if values is None:
                r_value = self.state.top(size * self.arch.byte_width)
            else:
                r_value = next(iter(next(iter(values.values.values()))))
            return RichR(r_value, variable=None, typevar=None)

        if not values:
            # the value does not exist. create a new variable
            variable = SimRegisterVariable(offset, size,
                                           ident=self.variable_manager[self.func_addr].next_variable_ident(
                                               'register'),
                                           region=self.func_addr,
                                           )
            value = self.state.top(size * self.state.arch.byte_width)
            value = self.state.annotate_with_variables(value, [(0, variable)])
            self.state.register_region.store(offset, value)
            self.variable_manager[self.func_addr].add_variable('register', offset, variable)

            value_list = [{ value }]
        else:
            value_list = list(values.values.values())

        variable_set = set()
        for value_set in value_list:
            for value in value_set:
                for _, var in self.state.extract_variables(value):
                    self.variable_manager[self.func_addr].read_from(var, None, codeloc, atom=expr)
                    variable_set.add(var)

        if self.arch.is_artificial_register(offset, size) or offset == self.arch.sp_offset:
            typevar = None
            var = None
        else:
            # we accept the precision loss here by only returning the first variable
            # FIXME: Multiple variables
            typevar = None
            var = None
            if variable_set:
                var = next(iter(variable_set))

                # add delayed type constraints
                if var in self.state.delayed_type_constraints:
                    for constraint in self.state.delayed_type_constraints[var]:
                        self.state.add_type_constraint(constraint)
                    self.state.delayed_type_constraints.pop(var)

                if var not in self.state.typevars:
                    typevar = typevars.TypeVariable()
                    self.state.typevars.add_type_variable(var, codeloc, typevar)
                else:
                    # FIXME: This is an extremely stupid hack. Fix it later.
                    # typevar = next(reversed(list(self.state.typevars[var].values())))
                    typevar = self.state.typevars[var]

        if len(value_list) == 1:
            r_value = next(iter(value_list[0]))
        else:
            r_value = self.state.top(size * self.arch.byte_width)  # fall back to top
        return RichR(r_value, variable=var, typevar=typevar)



import logging
import math

from ailment import Expr

from ... import AnalysesHub
from .engine_base import SimplifierAILEngine, SimplifierAILState
from .optimization_pass import OptimizationPass, OptimizationPassStage

_l = logging.getLogger(name=__name__)


class DivSimplifierAILEngine(SimplifierAILEngine):

    def _check_divisor(self, a, b, ndigits=6): #pylint: disable=no-self-use
        divisor_1 = 1 + (a//b)
        divisor_2 = int(round(a/float(b), ndigits))
        return divisor_1 if divisor_1 == divisor_2 else None

    def _ail_handle_Convert(self, expr):
        if expr.from_bits == 128 and expr.to_bits == 64:
            operand_expr = self._expr(expr.operand)
            if isinstance(operand_expr, Expr.BinaryOp) \
                    and operand_expr.op == 'Mul' \
                    and isinstance(operand_expr.operands[1], Expr.Const) \
                    and isinstance(operand_expr.operands[0], Expr.BinaryOp):
                if operand_expr.operands[0].op in {'Shr', 'DivMod'} \
                        and isinstance(operand_expr.operands[0].operands[1], Expr.Const):
                    if operand_expr.operands[0].op == 'Shr':
                        Y = operand_expr.operands[0].operands[1].value
                    else:
                        Y = int(math.log2(operand_expr.operands[0].operands[1].value))
                    C = operand_expr.operands[1].value
                    divisor = self._check_divisor(pow(2, 64+Y), C)
                    if divisor:
                        X = operand_expr.operands[0].operands[0]
                        new_const = Expr.Const(expr.idx, None, divisor, 64)
                        return Expr.BinaryOp(expr.idx, 'DivMod', [X, new_const], expr.signed, **expr.tags)

        return super()._ail_handle_Convert(expr)

    def _ail_handle_Shr(self, expr):

        operand_0 = self._expr(expr.operands[0])
        operand_1 = self._expr(expr.operands[1])

        X = None
        divisor = None

        if isinstance(operand_1, Expr.Const) \
            and isinstance(operand_0, Expr.BinaryOp) \
                and operand_0.op == 'DivMod' \
                    and isinstance(operand_0.operands[1], Expr.Const):
            divisor = operand_0.operands[1].value * pow(2, operand_1.value)
            X = operand_0.operands[0]
        if isinstance(operand_1, Expr.Const) \
            and isinstance(operand_0, Expr.Convert) \
                and isinstance(operand_0.operand, Expr.BinaryOp) \
                    and operand_0.operand.op == 'DivMod' \
                and isinstance(operand_0.operand.operands[1], Expr.Const):
            divisor = operand_0.operand.operands[1].value * pow(2, operand_1.value)
            X = operand_0.operand.operands[0]
        if isinstance(operand_1, Expr.Const) \
                and isinstance(operand_0, Expr.Convert) \
                and operand_0.from_bits == 128 \
                and operand_0.to_bits == 64:
            if isinstance(operand_0.operand, Expr.BinaryOp)\
                    and operand_0.operand.op == 'Mul':
                if isinstance(operand_0.operand.operands[1], Expr.Const):
                    C = operand_0.operand.operands[1].value
                    Y = operand_1.value
                    divisor = self._check_divisor(pow(2, 64+Y), C)
                    X = operand_0.operand.operands[0]
                elif isinstance(operand_0.operand.operands[0], Expr.BinaryOp) \
                        and operand_0.operand.operands[0].op in {'Shr', 'DivMod'}:
                    C = operand_0.operand.operands[1].value
                    Z = operand_1.value
                    if operand_0.operand.operands[0].op == 'Shr':
                        Y = operand_0.operand.operands[0].operands[1].value
                    else:
                        Y = int(math.log2(operand_0.operand.operands[0].operands[1].value))
                    divisor = self._check_divisor(pow(2, 64+Z+Y), C)
                    X = operand_0.operand.operands[0].operands[0]
        if isinstance(operand_1, Expr.Const) \
                and isinstance(operand_0, Expr.BinaryOp) \
                and operand_0.op == 'Add':
            add_0, add_1 = operand_0.operands
            Z = operand_1.value
            if add_0.has_atom(add_1) or add_1.has_atom(add_0):
                xC = add_1 if add_0.has_atom(add_1) else add_0
                x_xC = add_0 if add_0.has_atom(add_1) else add_1
                if isinstance(xC, Expr.Convert) and (xC.from_bits > xC.to_bits):
                    Y = xC.from_bits - xC.to_bits
                    if isinstance(xC.operand, Expr.BinaryOp) and xC.operand.op == 'Mul':
                        xC_ = xC.operand
                        if isinstance(xC_.operands[1], Expr.Const):
                            C = xC_.operands[1].value
                            X = xC_.operands[0]
                            if isinstance(x_xC, Expr.BinaryOp) and x_xC.op == 'Shr':
                                V_, V = x_xC.operands
                                if isinstance(V, Expr.Const):
                                    V = V.value
                                    if isinstance(V_, Expr.BinaryOp) and V_.op == 'Sub':
                                        if V_.operands[0] == X and V_.operands[1] == xC:
                                            divisor = self._check_divisor(pow(2, Y+V+Z), C*(pow(2, V) - 1) + pow(2, Y))
                # unsigned int here
                if isinstance(xC, Expr.BinaryOp) and xC.op == 'Mul':
                    if isinstance(xC.operands[1], Expr.Const) \
                        and isinstance(xC.operands[0], Expr.Convert):
                        C = xC.operands[1].value
                        X = xC.operands[0]
                        Y = X.from_bits - X.to_bits
                        if isinstance(x_xC, Expr.BinaryOp) and x_xC.op == 'Shr':
                            V_, V = x_xC.operands
                            if isinstance(V, Expr.Const):
                                V = V.value
                                if isinstance(V_, Expr.BinaryOp) and V_.op == 'Sub':
                                    if V_.operands[1] == xC:
                                        divisor = self._check_divisor(pow(2, Y+V+Z), C*(pow(2, V) - 1) + pow(2, Y))
                elif isinstance(xC, Expr.BinaryOp) and xC.op == 'Shr':
                    if isinstance(xC.operands[1], Expr.Const) \
                        and isinstance(xC.operands[0], Expr.BinaryOp) \
                            and xC.operands[0].op == 'Mul' \
                                and isinstance(xC.operands[0].operands[1], Expr.Const):
                        if isinstance(x_xC, Expr.BinaryOp) \
                            and isinstance(x_xC.operands[1], Expr.Const) \
                                and isinstance(x_xC.operands[0], Expr.BinaryOp) \
                                    and x_xC.op == 'Shr' and x_xC.operands[0].op == 'Sub':
                            X = xC.operands[0].operands[0]
                            C = xC.operands[0].operands[1].value
                            Y = xC.operands[1].value
                            V = x_xC.operands[1].value
                            if X == x_xC.operands[0].operands[0]:
                                divisor = self._check_divisor(pow(2, Y+V+Z), C*(pow(2, V) - 1) + pow(2, Y))

        # unsigned int
        if isinstance(operand_1, Expr.Const) \
            and isinstance(operand_0, Expr.BinaryOp) \
                and operand_0.op == 'Mul' \
                    and isinstance(operand_0.operands[1], Expr.Const):
            if isinstance(operand_0.operands[0], Expr.Convert):
                V = operand_0.operands[0].from_bits - operand_0.operands[0].to_bits
                C = operand_0.operands[1].value
                Z = operand_1.value
                X = operand_0.operands[0]
                divisor = self._check_divisor(pow(2, V+Z), C)
            elif isinstance(operand_0.operands[0], Expr.BinaryOp) \
                and isinstance(operand_0.operands[0].operands[1], Expr.Const) \
                    and operand_0.operands[0].op in {'Shr', 'DivMod'}:
                X = operand_0.operands[0].operands[0]
                V = 0
                ndigits = 6
                if isinstance(X, Expr.Convert):
                    V = X.from_bits - X.to_bits
                if V == 32:
                    ndigits = 5
                C = operand_0.operands[1].value
                Y = operand_0.operands[0].operands[1].value
                if operand_0.operands[0].op == 'DivMod':
                    Y = int(math.log2(operand_0.operands[0].operands[1].value))
                Z = operand_1.value
                divisor = self._check_divisor(pow(2, Y+Z+V), C, ndigits)
            else:
                X = operand_0.operands[0]
                Y = operand_1.value
                C = operand_0.operands[1].value
                divisor = self._check_divisor(pow(2, Y), C)

        if divisor and X:
            new_const = Expr.Const(expr.idx, None, divisor, 64)
            return Expr.BinaryOp(expr.idx, 'DivMod', [X, new_const], expr.signed, **expr.tags)


        if isinstance(operand_1, Expr.Const):
            if isinstance(operand_0, Expr.Register):
                new_operand = Expr.Const(operand_1.idx, None, 2**operand_1.value, operand_1.bits)
                return Expr.BinaryOp(expr.idx, 'DivMod', [operand_0, new_operand], expr.signed)
            elif isinstance(operand_0, Expr.BinaryOp) \
                and operand_0.op == 'Shr' \
                    and isinstance(operand_0.operands[1], Expr.Const):
                new_const = Expr.Const(operand_1.idx, None,
                    operand_0.operands[1].value+operand_1.value, operand_1.bits)
                return Expr.BinaryOp(expr.idx, 'Shr', [operand_0.operands[0], new_const], expr.signed, **expr.tags)

        if (operand_0, operand_1) != (expr.operands[0], expr.operands[1]):
            return Expr.BinaryOp(expr.idx, 'Shr', [operand_0, operand_1], expr.signed)
        return expr

    def _ail_handle_Mul(self, expr):

        operand_0, operand_1 = expr.operands

        if isinstance(operand_1, Expr.Const) \
            and isinstance(operand_0, Expr.BinaryOp) \
                and isinstance(operand_0.operands[1], Expr.Const) \
                    and operand_0.op in {'DivMod', 'Shr'}:
            if operand_0.op == 'DivMod':
                Y = int(math.log2(operand_0.operands[1].value))
            else:
                Y = operand_0.operands[1].value
            C = operand_1.value
            X = operand_0.operands[0]
            # there is a convert outside this expr
            V = 64
            if isinstance(X, Expr.Convert):
                V = X.from_bits - X.to_bits
            ndigits = 5 if V == 32 else 6
            if self._check_divisor(pow(2, V+Y), C, ndigits) and X:
                divisor = self._check_divisor(pow(2, Y+V), C, ndigits)
                new_const = Expr.Const(expr.idx, None, divisor, 64)
                return Expr.BinaryOp(expr.idx, 'DivMod', [X, new_const], expr.signed, **expr.tags)
        if isinstance(operand_1, Expr.Const) \
            and isinstance(operand_0, Expr.Convert) \
                and isinstance(operand_0.operand, Expr.BinaryOp) \
                    and isinstance(operand_0.operand.operands[1], Expr.Const) \
                and operand_0.operand.op in {'DivMod', 'Shr'}:
            if operand_0.operand.op == 'DivMod':
                Y = int(math.log2(operand_0.operand.operands[1].value))
            else:
                Y = operand_0.operand.operands[1].value
            C = operand_1.value
            X = operand_0.operand.operands[0]
            V = operand_0.from_bits - operand_0.to_bits
            ndigits = 5 if V == 32 else 6
            if self._check_divisor(pow(2, V+Y), C, ndigits) and X:
                divisor = self._check_divisor(pow(2, Y+V), C, ndigits)
                new_const = Expr.Const(expr.idx, None, divisor, 64)
                return Expr.BinaryOp(expr.idx, 'DivMod', [X, new_const], expr.signed, **expr.tags)
        return super()._ail_handle_Mul(expr)

    def _ail_handle_Div(self, expr):

        operand_0 = self._expr(expr.operands[0])
        operand_1 = self._expr(expr.operands[1])

        if isinstance(operand_1, Expr.Const) \
            and isinstance(operand_0, Expr.BinaryOp) \
            and operand_0.op in {'Div', 'DivMod'} \
            and isinstance(operand_0.operands[1], Expr.Const):

            new_const_value = operand_1.value * operand_0.operands[1].value
            new_const = Expr.Const(operand_1.idx, None, new_const_value, operand_1.bits)
            return Expr.BinaryOp(expr.idx, 'Div', [operand_0.operands[0], new_const], expr.signed, **expr.tags)

        if (operand_0, operand_1) != (expr.operands[0], expr.operands[1]):
            return Expr.BinaryOp(expr.idx, 'Div', [operand_0, operand_1], expr.signed, **expr.tags)
        return expr


class DivSimplifier(OptimizationPass):

    ARCHES = ["X86", "AMD64"]
    PLATFORMS = ["linux", "windows"]
    STAGE = OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION

    def __init__(self, func, **kwargs):

        super().__init__(func, **kwargs)

        self.state = SimplifierAILState(self.project.arch)
        self.engine = DivSimplifierAILEngine()

        self.analyze()

    def _check(self):
        return True, None

    def _analyze(self, cache=None):

        for block in list(self._graph.nodes()):
            new_block = block
            old_block = None

            while new_block != old_block:
                old_block = new_block
                new_block = self.engine.process(state=self.state.copy(), block=old_block.copy())
                _l.debug("new block: %s", new_block.statements)

            self._update_block(block, new_block)

AnalysesHub.register_default("DivSimplifier", DivSimplifier)
